@@ -5,16 +5,22 @@ package Mail::SPF::Query;
 #
 # 		       Meng Weng Wong
 #		  <mengwong+spf@pobox.com>
-# $Id: Query.pm,v 1.19 2003/11/29 21:38:39 devel Exp $
+# $Id: Query.pm,v 1.31 2003/12/19 03:20:53 devel Exp $
 # test an IP / sender address pair for pass/fail/nodata/error
 #
 # http://spf.pobox.com/
 #
-# this version is compatible with draft-02.9.txt
+# this version is compatible with draft-02.9.4.txt
 #
 # license: opensource.
 #
 # TODO: add ipv6 support
+#
+# BUGS:
+#  mengwong 20031211
+#    if there are multiple unrecognized mechanisms, they all
+#    need to be preserved in the 'unknown' Received-SPF header.
+#    right now only the first appears.
 # ----------------------------------------------------------
 
 use 5.006;
@@ -31,15 +37,15 @@ use Net::DNS qw(); # by default it exports mx, which we define.
 # 		       initialization
 # ----------------------------------------------------------
 
-my $GUESS_MECHS = "a/24 mx/24 ptr exists:%{d}.wl.trusted-forwarder.org exists:%{ir}.wl.trusted-forwarder.org";
+my $GUESS_MECHS = "a/24 mx/24 ptr exists:%{p}.wl.trusted-forwarder.org exists:%{ir}.wl.trusted-forwarder.org";
 
 my @KNOWN_MECHANISMS = qw( a mx ptr include ip4 ip6 exists all );
 
-my $MAX_RECURSION_DEPTH = 10;
+my $MAX_LOOKUP_COUNT    = 20;
 
 my $Domains_Queried = {};
 
-$VERSION = "1.9";
+$VERSION = "1.96";
 
 $CACHE_TIMEOUT = 120;
 
@@ -62,14 +68,15 @@ Mail::SPF::Query - query Sender Permitted From for an IP,email,helo
 
   if    ($result eq "pass")     { ... } # domain is not forged
   elsif ($result eq "fail")     { ... } # domain is forged
-  elsif ($result eq "softfail") { ... } # domain may be forged
   else                          {       # domain has not implemented SPF
     if    ($guess eq "pass")    { ... } # result based on $guess_mechs
     elsif ($guess eq "fail")    { ... } # result based on $guess_mechs
     else                        { ... } #
   }
 
-  The default $guess_mechs is "a/24 mx/24 ptr exists:%{d}.wl.trusted-forwarder.org".
+  The default $guess_mechs is "a/24 mx/24 ptr
+         exists:%{p}.wl.trusted-forwarder.org
+         exists:%{ir}.wl.trusted-forwarder.org".
 
 =head1 ABSTRACT
 
@@ -87,9 +94,10 @@ of an SMTP client IP.
                                            helo  =>"host.example.com") };
 
   optional parameters:
-     guess_mechs => "a/24 mx/24 ptr exists:%{d}.wl.trusted-forwarder.org",
+     guess_mechs => "a/24 mx/24 ptr exists:%{p}.wl.trusted-forwarder.org",
      debug => 1, debuglog => sub { print STDERR "@_\n" },
-     max_recursion_depth => 10,
+     max_lookup_count => 20,          # total number of SPF include/redirect queries
+     myhostname => "foo.example.com", # prepended to header_comment
 
   if ($@) { warn "bad input to Mail::SPF::Query: $@" }
 
@@ -111,18 +119,20 @@ sub new {
 
   $query->{sender} =~ s/<(.*)>/$1/g;
 
-  if (not $query->{helo}) { require Carp; import Carp qw(cluck); cluck ("Mail::SPF::Query: ->new() requires a \"helo\" argument.\n");
-			    $query->{helo} = "";
-			  }
-
-  $query->debuglog("new: ipv4=$query->{ipv4}, sender=$query->{sender}, helo=$query->{helo}");
-
   if (not ($query->{ipv4}   and length $query->{ipv4}))   { die "no IP address given to spfquery"   }
 
   for ($query->{sender}) { s/^\s+//; s/\s+$//; }
 
   ($query->{domain}) = $query->{sender} =~ /([^@]+)$/; # given foo@bar@baz.com, the domain is baz.com, not bar@baz.com.
 
+  if (not $query->{helo}) { require Carp; import Carp qw(cluck); cluck ("Mail::SPF::Query: ->new() requires a \"helo\" argument.\n");
+			    $query->{helo} = $query->{domain};
+			  }
+
+  $query->debuglog("new: ipv4=$query->{ipv4}, sender=$query->{sender}, helo=$query->{helo}");
+
+  ($query->{helo}) =~ s/.*\@//; # strip localpart from helo
+
   if (not $query->{domain}) {
     $query->debuglog("spfquery: sender $query->{sender} has no domain, using HELO domain $query->{helo} instead.");
     $query->{domain} = $query->{helo};
@@ -131,6 +141,8 @@ sub new {
 
   if (not length $query->{domain}) { die "unable to identify domain of sender $query->{sender}" }
 
+  $query->{orig_domain} = $query->{domain};
+
   $query->{loop_report} = [$query->{domain}];
 
   ($query->{localpart}) = $query->{sender} =~ /(.+)\@/;
@@ -141,6 +153,14 @@ sub new {
   $query->{Reversed_IP} = ($query->{ipv4} ? reverse_in_addr($query->{ipv4}) :
 			   $query->{ipv6} ? die "IPv6 not supported" : "");
 
+  if (not $query->{myhostname}) {
+    use Sys::Hostname; 
+    eval { require Sys::Hostname::Long };
+    $query->{myhostname} = $@ ? hostname() : Sys::Hostname::Long::hostname_long();
+  }
+
+  $query->{myhostname} ||= "localhost";
+
   $query->post_new(@_) if $class->can("post_new");
 
   return $query;
@@ -148,9 +168,9 @@ sub new {
 
 =head2 $query->result()
 
-  my ($result, $comment) = $query->result();
+  my ($result, $smtp_comment, $header_comment) = $query->result();
 
-C<$result> will be one of C<pass>, C<fail>, C<softfail>, C<unknown>, or C<error>.
+C<$result> will be one of C<pass>, C<fail>, C<unknown>, or C<error>.
 
 C<pass> means the client IP is a designated mailer for the
 sender.  The mail should be accepted subject to local policy
@@ -160,10 +180,6 @@ C<fail> means the client IP is not a designated mailer, and
 the sender wants you to reject the transaction for fear of
 forgery.
 
-C<softfail> means the transaction should be accepted but
-subject to further scrutiny because the domain is still
-transitioning toward SPF adoption.
-
 C<unknown> means the domain either does not publish SPF data
 or has a configuration error in the published data.
 
@@ -174,6 +190,10 @@ Results are cached internally for a default of 120 seconds.
 You can call C<-E<gt>result()> repeatedly; subsequent
 lookups won't hit your DNS.
 
+The smtp_comment should be displayed to the SMTP client.
+
+The header_comment goes into a Received-SPF header.
+
 =cut
 
 # ----------------------------------------------------------
@@ -188,9 +208,18 @@ sub result {
 
   # print STDERR "*** result = $result\n";
 
+  if ($result eq "fail") {
+    my $receiver = uri_escape($query->{myhostname});
+    my $why_url = $query->macro_substitute("http://spf.pobox.com/why.html?sender=%{S}&ip=%{I}&receiver=$receiver");
+    $smtp_comment ||= "please see $why_url";
+  }
+
   $query->{smtp_comment} = $smtp_comment;
 
-  return (lc $result, $smtp_comment, $query->header_comment($result)) if wantarray;
+  return (lc $result,
+	  $smtp_comment,
+	  "$query->{myhostname}: ". $query->header_comment($result)) if wantarray;
+
   return  lc $result;
 }
 
@@ -205,7 +234,7 @@ sub header_comment {
      : $result eq "fail"     ? "domain of $query->{sender} does not designate $ip as permitted sender"
      : $result eq "softfail" ? "transitioning domain of $query->{sender} does not designate $ip as permitted sender"
      : $result eq "error"    ? "error in processing during lookup of $query->{sender}"
-     : $result eq "UNKNOWN"  ? "an error occurred during SPF processing of domain of $query->{sender}"
+     : $result eq "UNKNOWN"  ? "unable to determine SPF status for $query->{sender}"
      : $result =~ /^UNKNOWN/ ? "encountered unrecognized mechanism during SPF processing of domain of $query->{sender}"
      :                         "domain of $query->{sender} does not designate permitted sender hosts" );
 
@@ -221,7 +250,7 @@ produce an educated guess anyway.
 It pretends the domain defined A, MX, and PTR mechanisms,
 plus a few others.  The default set of directives is
 
-  "a/24 mx/24 ptr exists:%{d}.wl.trusted-forwarder.org"
+  "a/24 mx/24 pt rexists:%{p}.wl.trusted-forwarder.org exists:%{ir}.wl.trusted-forwarder.org"
 
 That default set will return either "pass" or "unknown".
 
@@ -269,6 +298,7 @@ sub best_guess {
 				 );
 
   $guess_query->{depth} = 0;
+  $guess_query->top->{lookup_count} = 0;
 
   # if result is not defined, the domain has no SPF.
   #    perform fallback lookups.
@@ -332,6 +362,8 @@ sub spfquery {
   if ($query->{ipv4} and
       $query->{ipv4}=~ /^127\./) { return "pass", "localhost is always allowed." }
 
+  $query->top->{lookup_count}++;
+
   if ($query->is_looping)            { return "UNKNOWN", $query->is_looping }
   if ($query->can_use_cached_result) { return $query->cached_result; }
   else                               { $query->tell_cache_that_lookup_is_underway; }
@@ -375,8 +407,7 @@ sub spfquery {
 
     $query->debuglog("  executing redirect=$new_domain");
 
-    my $inner_query = $query->clone(sender => $query->{localpart} . '@' . $new_domain,
-				    domain => $new_domain,
+    my $inner_query = $query->clone(domain => $new_domain,
 				    depth  => $query->{depth} + 1,
 				    reason => "redirects to $new_domain",
 				   );
@@ -416,14 +447,18 @@ sub is_looping {
 	not defined $Domains_Queried->{$cache_point}->[0]);
 
   return (join " ", "exceeded maximum recursion depth:", @{$query->{loop_report}})
-    if ($query->{depth} >= $query->max_recursion_depth);
+    if ($query->{depth} >= $query->max_lookup_count);
+
+  return ("query caused more than " . $query->max_lookup_count . " lookups") if ($query->max_lookup_count 
+										 and
+										 $query->top->{lookup_count} > $query->max_lookup_count);
 
   return 0;
 }
 
-sub max_recursion_depth {
+sub max_lookup_count {
   my $query = shift;
-  return $query->{max_recursion_depth} || $MAX_RECURSION_DEPTH;
+  return $query->{max_lookup_count} || $MAX_LOOKUP_COUNT;
 }
 
 sub can_use_cached_result {
@@ -525,6 +560,7 @@ sub macro_substitute_item {
     $newval = $query->{localpart}       if (lc $field eq 'l');
     $newval = $query->{domain}          if (lc $field eq 'd');
     $newval = $query->{sender}          if (lc $field eq 's');
+    $newval = $query->{orig_domain}     if (lc $field eq 'o');
     $newval = $query->ip                if (lc $field eq 'i');
     $newval = $timestamp                if (lc $field eq 't');
     $newval = $query->{helo}            if (lc $field eq 'h');
@@ -591,6 +627,8 @@ sub evaluate_mechanism {
   if ({ map { $_=>1 } @KNOWN_MECHANISMS }->{$mechanism}) {
     my $mech_sub = "mech_$mechanism";
     my ($hit, $text) = $query->$mech_sub($query->macro_substitute($argument, 255));
+    no warnings 'uninitialized';
+    $query->debuglog("  evaluate_mechanism: $modifier$mechanism($argument) returned $hit $text");
 
     return if not $hit;
 
@@ -628,6 +666,15 @@ sub myquery {
 
   $query->debuglog("  myquery: doing $qtype query on $label");
 
+  for ($label) {
+    if (/\.\./ or /^\./) {
+      # convert .foo..com to foo.com, etc.
+      $query->debuglog("  myquery: fixing up invalid syntax in $label");
+      s/\.\.+/\./g;
+      s/^\.//;
+      $query->debuglog("  myquery: corrected label is $label");
+    }
+  }
   my $resquery = $query->resolver->query($label, $qtype);
 
   my $errorstring = $query->resolver->errorstring;
@@ -687,8 +734,7 @@ sub mech_include {
 
   $query->debuglog("  mechanism include: recursing into $argument");
 
-  my $inner_query = $query->clone(sender => $query->{localpart} . '@' . $argument,
-				  domain => $argument,
+  my $inner_query = $query->clone(domain => $argument,
 				  depth  => $query->{depth} + 1,
 				  reason => "includes $argument",
 				 );
@@ -750,7 +796,15 @@ sub mech_mx {
 
   my $domain_to_use = $argument || $query->{domain};
 
-  foreach my $mx ($query->myquery($domain_to_use, "MX", "exchange", "preference")) {
+  my @mxes = $query->myquery($domain_to_use, "MX", "exchange", "preference");
+
+  # if a domain has no MX record, we MUST NOT use its IP address instead.
+  # if (! @mxes) {
+  #   $query->debuglog("  mechanism mx: no MX found for $domain_to_use.  Will pretend it is its own MX, and test its IP address.");
+  #   @mxes = ($domain_to_use);
+  # }
+
+  foreach my $mx (@mxes) {
     # $query->debuglog("  mechanism mx: $mx");
 
     foreach my $a ($query->myquery($mx, "A", "address")) {
@@ -890,8 +944,8 @@ sub shorthand2value {
   my $shorthand = shift;
   return { "-" => "fail",
 	   "+" => "pass",
-	   "~" => "softfail",
-	   "?" => "unknown" } -> {$shorthand} || $shorthand;
+	   "~" => "UNKNOWN",
+	   "?" => "UNKNOWN" } -> {$shorthand} || $shorthand;
 }
 
 sub value2shorthand {
@@ -960,7 +1014,7 @@ sub interpolate_explanation {
 
     return if not $txt;
 
-    my $directive_set = bless { txt => $txt } , $class;
+    my $directive_set = bless { orig_txt => $txt, txt => $txt } , $class;
 
     TXT_RESPONSE:
     for ($txt) {
@@ -993,6 +1047,7 @@ sub interpolate_explanation {
 	if (my ($prefix, $lhs, $rhs) = $word =~ /^([-~+?]?)([\w_-]+)([\/:]\S*)?$/i) {
 	  $rhs =~ s/^://;
 	  $prefix ||= "+";
+	  $prefix = "?" if $prefix eq "~"; # softfail is deprecated, has become "unknown"
 	  $query->debuglog("  lookup:   TXT prefix=$prefix, lhs=$lhs, rhs=$rhs");
 	  push @{$directive_set->{mechanisms}}, [$prefix => lc $lhs => $rhs];
 	  next;
